// Copyright 2006 all rights reserved; see LICENSE file for BSD-style license

package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.util.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/** <font color=blue>a tree (or node in a tree); see jargon.txt for details</font> */
public class Tree<NodeType>
    extends PrintableTree<Tree<NodeType>>
    implements Iterable<Tree<NodeType>> {

    private final Input.Region     location;
    private final NodeType         ihead;
    private final Tree<NodeType>[] children;

    /** the number of children the tree has */
    public int               size() { return children.length; }

    /** the element at the head of the tree */
    public NodeType                 head()        { return ihead; }
    public NodeType              getHead()        { return ihead; }

    /** the tree's children */
    public Iterable<Tree<NodeType>> children()    { return this; }

    /** the tree's children */
    public Iterator<Tree<NodeType>> iterator()    { return new ArrayIterator(children); }

    /** get the <tt>i</t>th child */
    public Tree<NodeType> child(int i)            { return children[i]; }

    /** get the input region that this tree was parsed from */
    public Input.Region    getRegion() { return location; }

    public Tree(Input.Region loc, NodeType head)                            { this(loc, head, null); }
    public Tree(Input.Region loc, NodeType head, Tree<NodeType>[] children) { this(loc, head, children, false); }

    // FIXME: fairly inefficient because we keep copying arrays
    /** package-private constructor, allows setting the "lift" bit */
    Tree(Input.Region loc, NodeType head, Tree<NodeType>[] children, boolean lift) {
        this(loc, head, children, lift, false);
    }
    Tree(Input.Region loc, NodeType head, Tree<NodeType>[] children, boolean lift, boolean liftLeft) {
        this.location = loc;
        this.ihead = head;
        // FIXME: lift+liftLeft togheter
        if (liftLeft && children != null && children.length > 0) {
            Tree<NodeType> last = children[0];
            this.children = new Tree[(children.length-1)+last.children.length];
            System.arraycopy(children, 1, this.children, last.children.length, children.length-1);
            if (last.children.length > 0)
                System.arraycopy(last.children, 0, this.children, 0, last.children.length);
        } else if (lift && children != null && children.length > 0) {
            Tree<NodeType> last = children[children.length-1];
            this.children = new Tree[(children.length-1)+last.children.length];
            System.arraycopy(children, 0, this.children, 0, children.length-1);
            if (last.children.length > 0)
                System.arraycopy(last.children, 0, this.children, children.length-1, last.children.length);
        } else {
            this.children = ArrayUtil.clone(children, Tree.class);
        }
    }


    // PrintableTree /////////////////////////////////////////////////////////////////////////////

    protected String headToString() { return head()==null?null:head().toString(); }
    protected String headToJava()   {
      // FIXME
        if (head()==null) return null;
        if (head() instanceof ToJava) {
            StringBuffer sb = new StringBuffer();
            ((ToJava)head()).toJava(sb);
            return sb.toString();
        }
        return (head()==null?"null":("\""+StringUtil.toJavaString(head().toString())+"\""));
    }
    protected String left()   { return "{"; }
    protected String right()  { return "}"; }
    protected boolean ignoreSingleton() { return false; }


}
